@@ -25,16 +25,30 @@ from .graph import check_acyclic, get_pipeline, get_pipelines
 logger = logging.getLogger(__name__)
 
 
+@contextmanager
+def lock_repo(repo):
+    # pylint: disable=protected-access
+    depth = getattr(repo, "_lock_depth", 0)
+    repo._lock_depth = depth + 1
+
+    try:
+        if depth > 0:
+            yield
+        else:
+            with repo.lock, repo.state:
+                repo._reset()
+                yield
+                # Graph cache is no longer valid after we release the repo.lock
+                repo._reset()
+    finally:
+        repo._lock_depth = depth
+
+
 def locked(f):
     @wraps(f)
     def wrapper(repo, *args, **kwargs):
-        with repo.lock, repo.state:
-            # pylint: disable=protected-access
-            repo._reset()
-            ret = f(repo, *args, **kwargs)
-            # Our graph cache is no longer valid after we release the repo.lock
-            repo._reset()
-            return ret
+        with lock_repo(repo):
+            return f(repo, *args, **kwargs)
 
     return wrapper
 
@@ -44,11 +58,11 @@ class Repo:
 
     from dvc.repo.add import add
     from dvc.repo.brancher import brancher
-    from dvc.repo.checkout import _checkout
+    from dvc.repo.checkout import checkout
     from dvc.repo.commit import commit
     from dvc.repo.destroy import destroy
     from dvc.repo.diff import diff
-    from dvc.repo.fetch import _fetch
+    from dvc.repo.fetch import fetch
     from dvc.repo.freeze import freeze, unfreeze
     from dvc.repo.gc import gc
     from dvc.repo.get import get
@@ -66,7 +80,7 @@ class Repo:
     from dvc.repo.status import status
     from dvc.repo.update import update
 
-    def __init__(self, root_dir=None, scm=None, rev=None):
+    def __init__(self, root_dir=None, scm=None, rev=None, subrepos=False):
         from dvc.cache import Cache
         from dvc.data_cloud import DataCloud
         from dvc.lock import make_lock
@@ -116,6 +130,8 @@ class Repo:
             hardlink_lock=hardlink_lock,
             friendly=True,
         )
+        # used by RepoTree to determine if it should traverse subrepos
+        self.subrepos = subrepos
 
         self.cache = Cache(self)
         self.cloud = DataCloud(self)
@@ -581,19 +597,20 @@ class Repo:
         path_parts = os.path.normpath(path).split(os.path.sep)
         return self.DVC_DIR in path_parts
 
+    @cached_property
+    def repo_tree(self):
+        return RepoTree(self, subrepos=self.subrepos, fetch=True)
+
     @contextmanager
     def open_by_relpath(self, path, remote=None, mode="r", encoding=None):
         """Opens a specified resource as a file descriptor"""
 
-        tree = RepoTree(self, stream=True)
+        tree = RepoTree(self, stream=True, subrepos=True)
         path = os.path.join(self.root_dir, path)
         try:
             with self.state:
                 with tree.open(
-                    os.path.join(self.root_dir, path),
-                    mode=mode,
-                    encoding=encoding,
-                    remote=remote,
+                    path, mode=mode, encoding=encoding, remote=remote,
                 ) as fobj:
                     yield fobj
         except FileNotFoundError as exc:
@@ -604,14 +621,6 @@ class Repo:
     def close(self):
         self.scm.close()
 
-    @locked
-    def checkout(self, *args, **kwargs):
-        return self._checkout(*args, **kwargs)
-
-    @locked
-    def fetch(self, *args, **kwargs):
-        return self._fetch(*args, **kwargs)
-
     def _reset(self):
         self.__dict__.pop("graph", None)
         self.__dict__.pop("stages", None)
